;; the STICCS (the Scheme TIme Code Calculator: STICCS)
;
;;                    LICENSE AGREEMENT
;
; Copyright (c) 2004
;	Aaron Wen-yao Hsu.  All rights reserved.
;
; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions
; are met:
; 1. Redistributions of source code must retain the above copyright
;    notice, this list of conditions and the following disclaimer.
; 2. Redistributions in binary form must reproduce the above copyright
;    notice, this list of conditions and the following disclaimer in the
;    documentation and/or other materials provided with the distribution.
; 3. Neither the name of the author nor the names of its contributors
;    may be used to endorse or promote products derived from this software
;    without specific prior written permission.
;
; THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
; ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
; LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
; OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
; SUCH DAMAGE.
;

;; tc-add : list list integer boolean -> list
; Takes two timecodes and adds them together, using the boolean to 
; determine whether or not to use Drop Frame, and outputs the total.
;
; Example: (tc-add '(1 25 20 14) '(0 30 15 25) 30 #f) = '(1 55 36 9)
(define (tc-add tc1 tc2 fr df)
  (cond ((eqv? df #f) ; Using no Drop frames?
         (tc-add-carry ; Carry with initial addition
          (list (+ (car tc1) (car tc2)) ; Add hours
                (+ (cadr tc1) (cadr tc2)) ; Add minutes
                (+ (caddr tc1) (caddr tc2)) ; Add seconds
                (+ (cadddr tc1) (cadddr tc2))) ; Add frames
          fr))
        ((eqv? df #t) ; Using Drop Frames?
         (if (= fr 30) ; Check for proper framerate
             (tc-add ; Call addition with frames dealt with.
              (tc-add-df-calc (list ; Deal with frame drops
                           (car tc1) ; Set Hours
                           (cadr tc1) ; Set minutes
                           (caddr tc1) ; Set Seconds
                           ; Do initial frame addition
                           (+ (cadddr tc1) (cadddr tc2))))
              ; timecode2's frames have already been added, so 
              ; use a frame count of 0
              (list (car tc2) (cadr tc2) (caddr tc2) 0)
              30 #f) 
             #f))))

;; tc-add-df-calc : list -> list
; takes one list in timecode and cleans the frames 
; taking into account drop frames.
(define (tc-add-df-calc m)
  (cond ((> (caddr m) 59) ; Check for second carry
         (tc-add-df-calc (tc-add-carry-set! m 60 3)))
        ((> (cadr m) 59) ; Check for minute carry
         (tc-add-df-calc (tc-add-carry-set! m 60 2)))
        ((> (cadddr m) 29) ; Check for frame carry
         (if (or (= (cadr m) 0) (= (cadr m) 1)) ; Check for 00:|01:
             ; If at min 00 or 01, then carry normally
             (tc-add-df-calc (tc-add-carry-set! m 30 4))
             ; else add two additional frames
             (tc-add-df-calc (tc-add-carry-set! m 28 4))))
        (else m))) ; No more carries necessary, output timecode list

;; tc-add-carry : list integer -> list
; takes a list and carries all the timecode appropriately, ouputting 
; a list with the properly carried values
(define (tc-add-carry tc fr)
  (cond ((> (cadddr tc) (- fr 1)) ; Frames need carry
         (tc-add-carry (tc-add-carry-set! tc fr 4) fr))
        ((> (caddr tc) 59) ; Secs need carry
         (tc-add-carry (tc-add-carry-set! tc 60 3) fr))
        ((> (cadr tc) 59) ; Mins need carry
         (tc-add-carry (tc-add-carry-set! tc 60 2) fr))
        ((> (car tc) 99) ; Hour limit exceeded
         #f)
        (else tc))) ; No carrying necessary

;; tc-add-carry-set! : list integer integer -> list
; carries a select position in a given list according to a frame 
; rate
(define (tc-add-carry-set! tc fr pos)
  (cond ((= pos 4) ; Carry Frames
         (set-car! (cddr tc) (+ (caddr tc) 1)) ; Add one to Seconds
         ; Take 1 second worth of frames from the frames value
         (set-car! (cdddr tc) (- (cadddr tc) fr)))
        ((= pos 3) ; Carry Seconds
         (set-car! (cdr tc) (+ (cadr tc) 1)) ; Add one to Minutes
         ; Take 1 minute's worth of seconds from the seconds value
         (set-car! (cddr tc) (- (caddr tc) fr)))
        ((= pos 2) ; Carry Minutes
         (set-car! tc (+ (car tc) 1)) ; Add one to Hours
         ; Take 1 hour's worth of minutes from the minutes value
         (set-car! (cdr tc) (- (cadr tc) fr))))
  tc)

;; tc-mul : list integer fr df -> list
; takes a timecode and multiplies it x amount of times at given 
; framerate, with or without Drop Frames.
(define (tc-mul tc x fr df)
  ; We use a named let to enable a simple syntax up front, and then 
  ; use an additional parameter for correct recursion, so that we can 
  ; keep track of what "unit" we need to add each iteration.
;  (let mult ((tc tc) ; Set up environment for recursion
;             (tc2 tc) ; Use additional timecode for tracking
;             (x x)
;             (fr fr)
;             (df df))
;    (cond ((> x 1) ; Still need to add another unit
;           ; Call recursively with one addition and the "unit"
;           (mult (tc-add tc tc2 fr df) tc2 (- x 1) fr df))
;          (else tc)))) ; If no more addition is necessary, return list
  (do ((tcm (if (list? tc) tc x) (tc-add tcm (if (list? tc) tc x)
                                         fr df))
       (i (if (number? x) x tc) (- i 1)))
    ((= i 1) tcm)))

;; tc-sub : list list integer boolean -> list
; takes two timecodes with frame rate and subtracts the second from 
; the first. Allows for drop frames.
(define (tc-sub tc1 tc2 fr df)
  (cond ((eqv? df #f) ; Use Drop Frames? No.
         (tc-borrow (list ; Initial subtraction, call to borrow
                     (- (car tc1) (car tc2)) ; Sub hours
                     (- (cadr tc1) (cadr tc2)) ; Sub mins
                     (- (caddr tc1) (caddr tc2)) ; Sub secs
                     (- (cadddr tc1) (cadddr tc2))) ; Sub frames
                    fr))
        ((eqv? df #t) ; Use Drop frames? Yes.
         (if (= fr 30) ; Check for correct framerate
             (tc-sub ; Subtract again after drop frame adjustment
              (tc-sub-df-calc (list ; Perform drop frame adjustment
                               (car tc1) ; Set Hours
                               (cadr tc1) ; Set mins
                               (caddr tc1) ; set secs
                               ; Perform initial frame subtraction
                               (- (cadddr tc1) (cadddr tc2))))
              ; Since frame subtraction has already been done, the
              ; frames must be set to 0
              (list (car tc2) (cadr tc2) (caddr tc2) 0)
              30 #f)
             #f))))
  
;; tc-borrow : list fr -> list
; takes a timecode and performs borrowing based on framerate, returns 
; a timcecode format list.
(define (tc-borrow tc fr)
  (cond ((negative? (cadddr tc)) ; Frames need borrowing?
         (tc-borrow (tc-borrow-set! tc fr 4) fr))
        ((negative? (caddr tc)) ; Seconds need borrowing?
         (tc-borrow (tc-borrow-set! tc 60 3) fr))
        ((negative? (cadr tc)) ; Minutes need borrowing?
         (tc-borrow (tc-borrow-set! tc 60 2) fr))
        ((negative? (car tc)) ; Hours negative?
         #f) ; Go to lowest non-neg setting
        (else tc)))

;; tc-borrow-set! : list integer integer -> list
; Borrows from a position - 1 and adds to position according to 
; framerate, returns a timecode list
(define (tc-borrow-set! tc fr pos)
  (cond ((= pos 4) ; Borrow for frames?
         ; Remove one second from code
         (set-car! (cddr tc) (- (caddr tc) 1))
         ; Add one second worth of frames
         (set-car! (cdddr tc) (+ (cadddr tc) fr)))
        ((= pos 3) ; Borrow for seconds?
         ; Remove one minute
         (set-car! (cdr tc) (- (cadr tc) 1))
         ; Add one minute worth of seconds
         (set-car! (cddr tc) (+ (caddr tc) 60)))
        ((= pos 2) ; Borrow for minutes?
         ; Remove one hour
         (set-car! tc (- (car tc) 1))
         ; Add one hour's worth of minutes
         (set-car! (cdr tc) (+ (cadr tc) 60))))
  tc)

;; tc-sub-df-calc : list -> list
; Calculate drop frames on a timecode
(define (tc-sub-df-calc tc)
  (cond ((negative? (cadr tc)) ; Minutes need carry
         (tc-sub-df-calc (tc-borrow-set! tc 60 2)))
        ((negative? (caddr tc)) ; Seconds need carry
         (tc-sub-df-calc (tc-borrow-set! tc 60 3)))    
        ((negative? (cadddr tc)) ; Frames need carry
         (if (or (= (cadr tc) 0) (= (cadr tc) 1)) ; Check for 00:|01:
             ; Treat normally if at min 00 or 01
             (tc-sub-df-calc (tc-borrow-set! tc 30 4))
             ; else subtract two extra frames
             (tc-sub-df-calc (tc-borrow-set! tc 28 4))))
        (else tc))) ; No more carries necessary

;; tc-div : list list/integer integer boolean -> list/integer
; takes a timecode and either an integer or another timecode and 
; divides the second into the first, returning either a timecode or 
; an integer
(define (tc-div tc dv fr df)
  (cond ((list? dv)
         (do ((tcd tc (tc-sub tcd dv fr df))
              (i 0 (+ i 1)))
           ((eqv? (tc-sub tcd dv fr df) #f) i)))
        ((integer? dv)
         #f)))

;; get-tc : string -> list
; takes a timecode string and outputs a timecode list
(define (get-tc strg)
  (set! strg (string-append (make-string 
                             (- 8 (string-length strg))
                             #\0)
                            strg))
  (list (string->number (substring strg 0 2))
        (string->number (substring strg 2 4))
        (string->number (substring strg 4 6))
        (string->number (substring strg 6 8))))
   